require 'pry-janitor/version'
require 'singleton'
require 'pry'

class PryJanitor
  include Singleton

  CHECK_ACTIVITY_SECONDS = 60
  DEFAULT_TIMEOUT_SECONDS = 60 * 60 * 2
  PRY_HOOK_NAME = :pry_janitor

  ACTIVITY_SIGNAL = 'USR1'.freeze
  PAUSE_SIGNAL = 'USR2'.freeze
  KILL_SIGNAL = 'TERM'.freeze
  EXIT_SIGNAL = 'EXIT'.freeze

  def self.auto_exit_after(timeout = DEFAULT_TIMEOUT_SECONDS)
    instance.auto_exit_after(timeout)
  end

  def self.register_pry_hooks
    instance.register_pry_hooks
  end

  def auto_exit_after(timeout)
    kill_child
    with_install_handlers do
      @parent_pid = Process.pid
      @child_pid = Process.fork { child_process(timeout) }
    end
    kill_child_on_exit
  end

  def register_pry_hooks
    # pause last activity for long running commands
    Pry.hooks.add_hook(:before_eval, PRY_HOOK_NAME) do
      signal_child(PAUSE_SIGNAL)
    end

    # record activity when a command finishes, line is read, or session starts
    Pry.hooks.add_hook(:after_eval, PRY_HOOK_NAME) do
      signal_child(ACTIVITY_SIGNAL)
    end

    Pry.hooks.add_hook(:after_read, PRY_HOOK_NAME) do
      signal_child(ACTIVITY_SIGNAL)
    end

    Pry.hooks.add_hook(:before_session, PRY_HOOK_NAME) do
      signal_child(ACTIVITY_SIGNAL)
    end
  end

  private

  # 1. install signal handlers in the parent so the child is immediately ready
  # 2. re-install pre-existing signal handlers back into the parent
  def with_install_handlers
    prev_sig1 = Signal.trap(ACTIVITY_SIGNAL) { @last_activity = Time.now }
    prev_sig2 = Signal.trap(PAUSE_SIGNAL) { @last_activity = nil }
    yield
    Signal.trap(ACTIVITY_SIGNAL, prev_sig1)
    Signal.trap(PAUSE_SIGNAL, prev_sig2)
  end

  def child_process(timeout)
    loop do
      begin
        # kill parent and exit loop if last activity is longer than timeout
        if @last_activity && @last_activity < Time.now - timeout
          Process.kill(KILL_SIGNAL, @parent_pid)
        end
        sleep CHECK_ACTIVITY_SECONDS
      rescue Interrupt
        # parent exited, exit loop
        break
      end
    end
  end

  def signal_child(signal)
    Process.kill(signal, @child_pid) if @child_pid
  end

  def kill_child
    signal_child(KILL_SIGNAL)
  end

  def kill_child_on_exit
    Signal.trap(EXIT_SIGNAL) { kill_child }
  end
end

# need to register the hooks before pry starts
PryJanitor.register_pry_hooks

# auto exit if production rails console
if defined?(Rails)
  prod = Rails.env.production?
  console = Rails.const_defined?('Console')
  PryJanitor.auto_exit_after if prod && console
end
